import importlib
import os
import sys
import ctypes
import io
from vergeml.utils import VergeMLError
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'

class Library:
    @staticmethod
    def is_installed():
        raise NotImplementedError

    @staticmethod
    def version():
        raise NotImplementedError

    @staticmethod
    def setup(env):
        raise NotImplementedError

class KerasLibrary(Library):
    @staticmethod
    def is_installed():
        return bool(importlib.util.find_spec('keras'))

    @staticmethod
    def version():
        stderr = sys.stderr
        sys.stderr = open(os.devnull, 'w')
        import keras # pylint: disable=E0401
        sys.stderr = stderr
        return keras.__version__

    @staticmethod
    def setup(env):
        stderr = sys.stderr
        sys.stderr = open(os.devnull, "w")
        # pylint: disable=W0612
        try:
            import keras
        except Exception as e:
            raise e
        finally:
            sys.stderr = stderr

        from keras import backend as K
        if K.backend() == 'tensorflow':
            TensorFlowLibrary.setup(env)
            K.set_session(TensorFlowLibrary.create_session(env))

    @staticmethod
    def callback(env, display_progress, stats):
        from keras.callbacks import Callback

        class KerasCallback(Callback):

            def __init__(self, env, display_progress, stats):
                self.env = env
                self.display_progress = display_progress
                self.stats = stats
                self.callback = None
                self.current_epoch = 0
                self.current_step = 0

            def on_train_begin(self, logs=None):
                logs = KerasCallback._xform_logs(logs)
                self.callback = env.progress_callback(self.params['epochs'], self.params['steps'],
                                                      self.display_progress, self.stats)

            def on_train_end(self, logs=None):
                logs = KerasCallback._xform_logs(logs)
                self.callback(self.current_epoch, self.current_step, **logs)

            def on_epoch_begin(self, epoch, logs=None):
                logs = KerasCallback._xform_logs(logs)
                self.callback(self.current_epoch, self.current_step, **logs)

            def on_epoch_end(self, epoch, logs=None):
                logs = KerasCallback._xform_logs(logs)
                self.current_epoch += 1
                self.callback(self.current_epoch, self.current_step, **logs)

            def on_batch_begin(self, batch, logs=None):
                logs = KerasCallback._xform_logs(logs)
                self.callback(self.current_epoch, self.current_step, **logs)

            def on_batch_end(self, batch, logs=None):
                logs = KerasCallback._xform_logs(logs)
                self.current_step += 1
                self.callback(self.current_epoch, self.current_step, **logs)

            @staticmethod
            def _xform_logs(logs):
                from copy import deepcopy
                logs = deepcopy(logs or {})
                for k in ('size', 'batch', 'epoch'):
                    if k in logs:
                        del logs[k]
                return {k.replace('_', '-'):v for k, v in logs.items()}

        return KerasCallback(env, display_progress, stats)

class TensorFlowLibrary(Library):
    @staticmethod
    def is_installed():
        return bool(importlib.util.find_spec('tensorflow'))

    @staticmethod
    def version():
        import tensorflow # pylint: disable=E0401
        tensorflow.logging.set_verbosity(tensorflow.logging.ERROR)
        return tensorflow.__version__ #pylint: disable=E1101

    @staticmethod
    def create_session(env):
        TensorFlowLibrary.setup(env)
        import tensorflow as tf # pylint: disable=E0401

        devid = env.get('device.id')
        if devid != "auto":
            if devid == "cpu":
                devnum = None
            else:
                devid, devnum = devid.split(":")
        else:
            devid = None
            devnum = None

        args = {}
        if devid == "cpu":
            args['device_count'] = {'GPU': 0}

        config = tf.ConfigProto(**args)
        if devid == "gpu":
            config.gpu_options.visible_device_list = devnum  # pylint: disable=locally-disabled, E1101

        device_memory = env.get('device.memory')
        if device_memory != 'auto':
            # pylint: disable=E1101
            try:
                fraction = float(device_memory.rstrip("%"))/100.
                config.gpu_options.per_process_gpu_memory_fraction = fraction
            except Exception:
                raise VergeMLError(f'Invalid value for device.memory: {device_memory}.',
                                    'Please use a percentage value, e.g. 50%.', hint_type='value', hint_key='device.memory')
            # pylint: enable=E1101

        if env.get('device.grow-memory'):
            config.gpu_options.allow_growth = True  # pylint: disable=locally-disabled, E1101

        return tf.Session(config=config)

    @staticmethod
    def setup(env):
        import tensorflow # pylint: disable=E0401
        tensorflow.logging.set_verbosity(tensorflow.logging.ERROR)
        tensorflow.set_random_seed(env.get('random-seed'))


class TorchLibrary(Library):
    @staticmethod
    def is_installed():
        return bool(importlib.util.find_spec('torch'))

    @staticmethod
    def version():
        import torch # pylint: disable=E0401
        return torch.__version__

    @staticmethod
    def setup(env):
        import torch # pylint: disable=E0401
        import torch.cuda # pylint: disable=E0401
        torch.manual_seed(env.get('random-seed'))
        torch.cuda.manual_seed(env.get('random-seed'))

class NumPyLibrary(Library):
    @staticmethod
    def is_installed():
        return bool(importlib.util.find_spec('numpy'))

    @staticmethod
    def version():
        import numpy # pylint: disable=E0401
        return numpy.__version__

    @staticmethod
    def setup(env):
        import numpy
        numpy.random.seed(env.get('random-seed'))

class PythonInterpreter(Library):
    @staticmethod
    def is_installed():
        return True

    @staticmethod
    def version():
        return ".".join(map(str, sys.version_info[:3]))

    @staticmethod
    def setup(env):
        import random
        random.seed(env.get('random-seed'))

_CUDA_SUCCESS = 0
_CUDA_OBJECT = None
_HAS_CUDA = None

class CudaLibrary(Library):

    @staticmethod
    def is_installed():
        CudaLibrary._setup_cuda_object()
        return _HAS_CUDA

    @staticmethod
    def version():
        cuda = _CUDA_OBJECT
        version = ctypes.c_int()
        if cuda.cuDriverGetVersion(ctypes.byref(version)) == _CUDA_SUCCESS:
            val = version.value
            maj = int(val/1000)
            min = int((val-(maj*1000))/10)
            return f"{maj}.{min}"
        else:
            return "Unknown"

    @staticmethod
    def devices():

        cuda = _CUDA_OBJECT
        CU_DEVICE_ATTRIBUTE_MULTIPROCESSOR_COUNT = 16
        CU_DEVICE_ATTRIBUTE_MAX_THREADS_PER_MULTIPROCESSOR = 39
        CU_DEVICE_ATTRIBUTE_CLOCK_RATE = 13
        CU_DEVICE_ATTRIBUTE_MEMORY_CLOCK_RATE = 36

        def ConvertSMVer2Cores(major, minor):
            # Returns the number of CUDA cores per multiprocessor for a given
            # Compute Capability version. There is no way to retrieve that via
            # the API, so it needs to be hard-coded.
            return {(1, 0): 8,
                    (1, 1): 8,
                    (1, 2): 8,
                    (1, 3): 8,
                    (2, 0): 32,
                    (2, 1): 48,
            }.get((major, minor), 192) # 3.0 and above

        nGpus = ctypes.c_int()
        name = b' ' * 100
        cc_major = ctypes.c_int()
        cc_minor = ctypes.c_int()
        cores = ctypes.c_int()
        threads_per_core = ctypes.c_int()
        clockrate = ctypes.c_int()
        freeMem = ctypes.c_size_t()
        totalMem = ctypes.c_size_t()

        result = ctypes.c_int()
        device = ctypes.c_int()
        context = ctypes.c_void_p()

        res = []

        if cuda.cuDeviceGetCount(ctypes.byref(nGpus)) != _CUDA_SUCCESS:
            return None

        for i in range(nGpus.value):
            dev = {}
            if cuda.cuDeviceGet(ctypes.byref(device), i) != _CUDA_SUCCESS:
                return None
            dev['id'] = i

            if cuda.cuDeviceGetName(ctypes.c_char_p(name), len(name), device) == _CUDA_SUCCESS:
                dev['name'] = name.split(b'\0', 1)[0].decode()
            if cuda.cuDeviceComputeCapability(ctypes.byref(cc_major), ctypes.byref(cc_minor), device) == _CUDA_SUCCESS:
                dev['compute_capability'] = (cc_major.value, cc_minor.value)
            if cuda.cuDeviceGetAttribute(ctypes.byref(cores), CU_DEVICE_ATTRIBUTE_MULTIPROCESSOR_COUNT, device) == _CUDA_SUCCESS:
                dev['multiprocessors'] = cores.value
                dev['cuda_cores'] = cores.value * ConvertSMVer2Cores(cc_major.value, cc_minor.value)
                if cuda.cuDeviceGetAttribute(ctypes.byref(threads_per_core), CU_DEVICE_ATTRIBUTE_MAX_THREADS_PER_MULTIPROCESSOR, device) == _CUDA_SUCCESS:
                    dev['concurrent_threads'] = (cores.value * threads_per_core.value)
            if cuda.cuDeviceGetAttribute(ctypes.byref(clockrate), CU_DEVICE_ATTRIBUTE_CLOCK_RATE, device) == _CUDA_SUCCESS:
                dev['gpu_clock_mhz'] = clockrate.value / 1000.
            if cuda.cuDeviceGetAttribute(ctypes.byref(clockrate), CU_DEVICE_ATTRIBUTE_MEMORY_CLOCK_RATE, device) == _CUDA_SUCCESS:
                dev['memory_clock_mhz'] = clockrate.value / 1000.
            result = cuda.cuCtxCreate(ctypes.byref(context), 0, device)
            if result != _CUDA_SUCCESS:
                return None
            else:
                try:
                    if cuda.cuMemGetInfo(ctypes.byref(freeMem), ctypes.byref(totalMem)) == _CUDA_SUCCESS:
                        dev['total_memory_mib'] = totalMem.value / 1024**2
                        dev['free_memory_mib'] = freeMem.value / 1024**2
                    else:
                        return None
                finally:
                    cuda.cuCtxDetach(context)
            res.append(dev)
        return res

    @staticmethod
    def devices_info():
        res = []
        devices = CudaLibrary.devices()
        for device in devices:
            buffer = io.StringIO()
            print("Device ID:            ", device['id'], file=buffer)
            if 'name' in device:
                print("Device Name:          ", device['name'], file=buffer)
            if 'compute_capability' in device:
                print("Compute Capability:   ", "%d.%d" % device['compute_capability'], file=buffer)
            if 'multiprocessors' in device:
                print("Multiprocessors:      ", device['multiprocessors'], file=buffer)
            if 'cuda_cores' in device:
                print("CUDA Cores:           ", device['cuda_cores'], file=buffer)
            if 'concurrent_threads' in device:
                print("Concurrent Threads:   ", device['concurrent_threads'], file=buffer)
            if 'gpu_clock_mhz' in device:
                print("GPU Clock:            ", "%g MHz" % device['gpu_clock_mhz'], file=buffer)
            if 'memory_clock_mhz' in device:
                print("Memory Clock:         ", "%g MHz" % device['memory_clock_mhz'], file=buffer)
            if 'total_memory_mib' in device:
                print("Total Memory:         ", "%ld MiB" % device['total_memory_mib'], file=buffer)
            if 'free_memory_mib' in device:
                print("Free Memory:          ", "%ld MiB" % device['free_memory_mib'], file=buffer)
            res.append(buffer.getvalue())
        return res

    @staticmethod
    def setup(env):
        pass

    # Code from:
    # https://gist.github.com/f0k/63a664160d016a491b2cbea15913d549

    @staticmethod
    def _setup_cuda_object():
        global _CUDA_OBJECT
        global _HAS_CUDA

        libnames = ('libcuda.so', 'libcuda.dylib', 'nvcuda.dll')
        for libname in libnames:
            try:
                cuda = ctypes.CDLL(libname)
                if cuda.cuInit(0) == _CUDA_SUCCESS:
                    _CUDA_OBJECT = cuda
                    _HAS_CUDA = True
                else:
                    _CUDA_OBJECT = None
                    _HAS_CUDA = False
            except OSError:
                continue
            else:
                break
        else:
            _CUDA_OBJECT = None
            _HAS_CUDA = False




from vergeml.command import command, CommandPlugin
from vergeml.option import option
from vergeml.utils import VergeMLError
import os
import os.path
import importlib
from vergeml.wsgi import WSGIApp
import sys
import webbrowser
import queue
from vergeml.display import DISPLAY

# Don't output text
DISPLAY.quiet = True

@command('rest', descr="Start a REST Server.")
@option('@AI')
@option('host', type=str, default='0.0.0.0', descr="Which host to listen on.")
@option('port', type=int, default=2204, descr="Which port to use.")
@option('no-browser', type=bool, default=False, descr="Don't open the browser.", flag=True)

class RestService(CommandPlugin):

    def __call__(self, args, env):
        if not bool(importlib.util.find_spec('waitress')):
            raise VergeMLError("Package waitress is not installed.",
                               "To run a REST server, please install waitress first.")

        from waitress.server import create_server
        print("Loading @{} ...".format(args['@AI']))

        assert env.model_plugin
        wsgiapp =  WSGIApp(env)
        server = create_server(wsgiapp.handler, host=args['host'], port=args['port'], threads=1)
        q = queue.Queue()

        class LoadAIOnHandlerThread:

            def service(self):
                try:
                    env.model_plugin.load(env)
                    q.put("DONE")
                except Exception as e:
                    q.put(e)

            def defer(self):
                pass

            def cancel(self):
                pass

        server.task_dispatcher.add_task(LoadAIOnHandlerThread())
        res = q.get()
        if isinstance(res, Exception):
            raise res
        else:
            assert res == "DONE"

        url = "http://{}:{}".format(server.effective_host, server.effective_port)

        print("Serving on " + url)
        if not args['no-browser']:
            webbrowser.open(url)

        server.run()


from vergeml.utils import VergeMLError
import os.path
import csv
import numpy as np

def load_labels(env):
    path = os.path.join(env.checkpoints_dir(), "labels.txt")

    if not os.path.exists(path):
        raise FileExistsError(path)
    
    with open(path) as labelsfile:
        labels = labelsfile.read().strip().splitlines()
        return labels

def load_predictions(env, nclasses):
    path = os.path.join(env.stats_dir(), "predictions.csv")

    if not os.path.exists(path):
        raise FileExistsError(path)

    with open(path, newline='') as csvfile:
        y_score = []
        y_test = []
        csv_reader = csv.reader(csvfile, dialect="excel")
        for row in csv_reader:
            assert len(row) == nclasses * 2
            y_score.append(list(map(float, row[:nclasses])))
            y_test.append(list(map(float, row[nclasses:])))
        
        y_score = np.array(y_score)
        y_test = np.array(y_test)

        return y_test, y_score
    


from vergeml.utils import parse_split, VergeMLError
import random
from typing import Optional, Any, List, Tuple
import hashlib
import glob
import os
import os.path
import json
from operator import methodcaller
import io
from vergeml.option import Option
from vergeml.plugins import PLUGINS
from copy import deepcopy
from functools import reduce


class Sample:
    """The datastructure used for loading and manipulating samples"""

    def __init__(self, x: Any, y: Any, meta: dict, rng: Optional[random.Random]):
        """
        :param x: the sample data
        :param y: the ground truth
        :param meta: sample metadata, i.e. the split, the source filename
        :param rng: a per sample random generator
        """
        self.x = x
        self.y = y
        self.meta = meta
        self.rng = rng

class SourcePlugin:
    """The base class for getting sample data.

    Whenever you request data from vergeml.Data.load(), it will use a subclass of
    SourcePlugin in the background and call its read_sample() and num_samples() methods.

    ```SourcePlugin``` takes a number of configuration parameters and sets up default values in case they are missing:

    | Parameter          | Default       | YAML               | Command Line   |
    | ------------------ | ------------- | ------------------ | -------------- |
    | samples_dir        | samples       | samples-dir        | --samples-dir  |
    | cache_dir          | .cache        | cache-dir          | --cache-dir    |
    | random_seed        | 42            | random-seed        | --random-seed  |
    | trainings_dir      | ./trainings   | N/A                | N/A            |
    | val_dir            | None          | val-split  [2]     | --val          |
    | val_num            | None          | val-split  [2]     | --val          |
    | val_perc           | 10 [1]        | val-split  [2]     | --val          |
    | test_dir           | None          | test-split [2]     | --test         |
    | test_num           | None          | test-split [2]     | --test         |
    | test_perc          | 10 [1]        | test-split [2]     | --test         |
    | input_patterns     | */**          | input-patterns [3] | N/A            |

    [1] val_perc and test_perc are set to 10 only when the corresponding _dir and _val values are not set.

    [2] val-split and test-split are interpreted as follows: if the value ends with a percentage sign,
        the corresponding _perc value is set. if the value is numeric, it will set up the corresponding _num value.
        Otherwise, it is interpreted as a path.

        For example:
        val-split=10% - sets val_perc to 10
        val-split=100 - sets val_num to 100
        val-split=/data/imgs - sets val_dir to '/data/imgs'

    [3] input-patterns is available to be set in YAML file when the SourcePlugin subclass defines input_patterns
        via @source
    """


    def __init__(self, args: dict={}):
        self.meta = {}
        self.args = deepcopy(args)

        # since input-patterns can be used in many different kinds of io sources, grab it too.
        self.input_patterns = args.get('input-patterns', '**/*')
        if isinstance(self.input_patterns, str):
            self.input_patterns = self.input_patterns.split(",")
        self.samples_dir = args.get('samples-dir', 'samples')
        self.cache_dir = args.get('cache-dir', '.cache')
        self.random_seed = args.get('random-seed', 42)
        self.trainings_dir = args.get('trainings-dir', './trainings')

        self._cached_file_state = None

        spltype, splval = parse_split(args.get('val-split', '10%'))
        self.val_dir = splval if spltype == 'dir' else None
        self.val_num = splval if spltype == 'num' else None
        self.val_perc = splval if spltype == 'perc' else None

        spltype, splval = parse_split(args.get('test-split', '10%'))
        self.test_dir = splval if spltype == 'dir' else None
        self.test_num = splval if spltype == 'num' else None
        self.test_perc = splval if spltype == 'perc' else None

    def begin_read_samples(self):
        """Called when the system starts reading samples"""

    def num_samples(self, split: str) -> int:
        """Returns the total number of samples available in the split."""
        raise NotImplementedError

    def read_samples(self, split: str, index: int, n: int=1) -> Sample:
        """Read a sample.

        :param split: a split ("train", "val" or "test")
        :param index: sample index
        :param n: how many samples to read

        The job of this method is to find the sample represented by the split and
        index and return the sample."""
        raise NotImplementedError

    def read_raw_samples(self, split: str, index: int, n: int=1) -> tuple:
        """Read raw sample data.

        This is used for caching. Override this to provide a more compact representation,
        e.g. image bytes
        """
        return self.read_samples(split, index, n)

    def recover_raw_sample(self, sample) -> Sample:
        """Recover a sample from raw sample data.

        Used for caching. Recover the value from read_raw_samples
        """
        return sample

    def end_read_samples(self):
        """Called when the system is done reading samples.

        NOTE: Under some circumstances this method is never called, for example when the source
        is used with infinite iterators or when an iterator is not consumed until the end.

        TODO: is this still true?"""

    def transform(self, sample):
        """Return the sample with x and y transformed to its final form."""
        raise NotImplementedError

    def output_shape(self):
        """Return the output shape after transform or None"""
        return None

    def hash(self, state: str) -> str:
        """Generate a hash representing the current sample state.

        :param state: string capturing the current system configuration state

        This function must be overridden to generate a meaningful hash for the current
        set of input samples."""

        newstate = io.BytesIO(state.encode('utf-8'))

        for k in ('input_patterns', 'samples_dir', 'val_dir', 'val_num', 'val_perc',
                  'test_dir', 'test_num', 'test_perc', 'random_seed'):
            newstate.write(str(getattr(self, k)).encode('utf-8'))

        md5 = hashlib.md5()
        md5.update(newstate.getvalue())
        return md5.hexdigest()

    def split(self, num_samples: int):
        """Split the dataset in train, val and test sets by percentage or absolute count.

        It works by receiving the total number of samples and a configuration
        object, and calculates an array of indices per split.:

        :param num_samples: the total number of samples

        :return: a tuple of indices for (train, val, test)
        """

        val_num = int(num_samples * self.val_perc // 100) if self.val_perc else self.val_num or 0
        test_num = int(num_samples * self.test_perc // 100) if self.test_perc else self.test_num or 0

        if val_num + test_num > num_samples:
            hint_key = None
            hint_type = None

            if self.val_num:
                hint_key = 'val'
                hint_type = 'val-split'
            elif self.test_num:
                hint_key = 'val'
                hint_type = 'test-split'

            raise VergeMLError("There are not enough samples to provide the configured number for the val and test split",
                               "If you use absolute numbers for 'val-split' or 'test-split', try to lower them",
                               help_topic='split', hint_key=hint_key, hint_type=hint_type)

        rng = random.Random(self.random_seed)
        indices = rng.sample(range(num_samples), num_samples)
        val, test, train = indices[:val_num], indices[val_num:val_num + test_num], indices[val_num + test_num:]
        return train, val, test

    def scan(self, path, exclude=[]) -> List[str]:
        """Scan path for matching files.

        :param path: the path to scan
        :param exclude: a list of directories to exclude

        :return: a list of sorted filenames
        """
        res = []
        path = path.rstrip("/").rstrip("\\")
        for pat in self.input_patterns:
            res.extend(glob.glob(path + os.sep + pat, recursive=True))

        res = list(filter(lambda p: os.path.isfile(p), res))

        if exclude:
            def excluded(path):
                for e in exclude:
                    if path.startswith(e):
                        return True
                return False

            res = list(filter(lambda p: not excluded(p), res))

        return sorted(res)

    def scan_dirs(self) -> Tuple[List[str], List[str], List[str]]:
        """Scan directories for matching files.

        :return: a tuple of files in samples_dir, val_dir, test_dir matching input_patterns
        """
        exclude = list(filter(None, (self.val_dir, self.test_dir)))
        train = self.scan(self.samples_dir, exclude)
        val = self.scan(self.val_dir) if self.val_dir else []
        test = self.scan(self.test_dir) if self.test_dir else []
        return train, val, test

    def normalize_filename(self, split, filename):
        """Return the filename without the parent samples directory
        """
        directory = self.samples_dir
        if split == 'val' and self.val_dir:
            directory = self.val_dir
        elif split == 'test' and self.test_dir:
            directory = self.test_dir

        return filename[len(directory):].strip(os.sep)


    def scan_and_split_files(self, files = None):
        """Return a dictionary of train, val and test files.
        """

        train_files, val_files, test_files = files or self.scan_dirs()

        def fromidx(idx):
            return [train_files[i] for i in idx]

        strain, sval, stest = self.split(len(train_files))

        def makemeta(split):
            return lambda filename: dict(split=split, filename=self.normalize_filename(split, filename))

        train = fromidx(strain)
        train_meta = map(makemeta('train'), train)
        val = val_files or fromidx(sval)
        val_meta = map(makemeta('val'), val)
        test = test_files or fromidx(stest)
        test_meta = map(makemeta('test'), test)

        return dict(
            train=list(zip(train, train_meta)),
            val=list(zip(val, val_meta)),
            test=list(zip(test, test_meta)))

    def hash_files(self, files):
        """A default implementation for hash based on files
        """
        if not self._cached_file_state:
            self._cached_file_state = io.BytesIO()

            for (split, files) in files.items():
                for path, _ in files:
                    fstate = "{}{}{}{}".format(split, path, os.path.getmtime(path), os.path.getsize(path))
                    self._cached_file_state.write(fstate.encode('utf-8'))

        return self._cached_file_state.getvalue().decode("utf-8")

    def options(self):
        return Option.discover(self)


    def configuration(self):
        """Return the configuration of the Source instance.

        Used for calculating the hash value when caching.
        """
        return self.args

    def supports_preview(self):
        return False

    _COUNTER = {}

    def begin_preview(self, output_dir: str):
        """Called when the system starts previewing samples"""

    def write_preview(self, output_dir: str, split: str, sample: Sample):
        raise NotImplementedError

    def end_preview(self, output_dir: str):
        """Called when the system is done previewing samples"""

    def preview_filename(self, path):
        """Generate a filename for previews, appending a number when the file already exists.
        """
        if not os.path.exists(path):
            return path

        pcounter = SourcePlugin._COUNTER
        if not path in pcounter:
            pcounter[path] = 1
        fname, ext = os.path.splitext(path)
        counter = pcounter[path]

        while os.path.exists("{}_{}{}".format(fname, counter, ext)):
            counter += 1
        pcounter[path] = counter

        return "{}_{}{}".format(fname, counter, ext)


_SOURCE_META_KEY = '__vergeml_source__'
def source(name, descr=None, long_descr=None, input_patterns=None):
    """Define a source.

    :param name:            Name of the source.
    :param descr:           A short description of the source
    :param long_descr:      A long description
    :param input_patterns:  When the source takes input patterns as argument, a single pattern
                            or a list of pattenrs. Otherwise None.
    """
    def decorator(o):
        # if getattr(o, _SOURCE_META_KEY, None):
        #     print(getattr(o, _SOURCE_META_KEY, None).name, name)
        # assert getattr(o, _SOURCE_META_KEY, None) is None

        options = Option.discover(o)
        if input_patterns:
            assert isinstance(input_patterns, (str, list))
            input_patterns_option = Option('input-patterns', input_patterns, type='Union[str, List[str]]',
                                            descr="Controls which files are loaded.",
                                            transform=lambda v: v if isinstance(v, list) else v.split(","))
            options.append(input_patterns_option)

        cmd = Source(name,
                    descr=descr,
                    long_descr=long_descr,
                    options=options)
        setattr(o, _SOURCE_META_KEY, cmd)
        return o
    return decorator

class Source:
    def __init__(self, name, descr=None, long_descr=None, options=None, plugins=PLUGINS):
        self.name = name
        self.descr = descr
        self.long_descr = long_descr
        self.options = options or []
        self.plugins = plugins

    @staticmethod
    def discover(o, plugins=PLUGINS):
        res = None
        if hasattr(o, _SOURCE_META_KEY):
            res = getattr(o, _SOURCE_META_KEY)
            res.plugins = plugins
            for option in res.options:
                option.plugins = plugins
        return res



from vergeml.img import ImageType, RESIZE_METHODS, RESIZE_MODES, resize_image
from vergeml.operation import OperationPlugin, operation
from vergeml.option import option
from PIL import Image

@operation('resize', topic="image", descr="Resize an image to a fixed size.")
@option('width', type=int, descr="Width of the new size.", validate='>0')
@option('height', type=int, descr="Height of the new size.", validate='>0')
@option('channels', type=int, descr="Number of channels.", validate=(0, 3))
@option('method', type=str, descr="Scaling Method.", default="antialias", validate=RESIZE_METHODS)
@option('mode', type=str, descr="Scaling Mode.", default="fill", validate=RESIZE_MODES)
class ResizeOperation(OperationPlugin):
    type = ImageType
    

    def __init__(self, width, height, channels=None, method='antialias', mode='fill', apply=None):
        assert method in RESIZE_METHODS
        assert mode in RESIZE_MODES
        assert channels in (None, 1, 3)

        super().__init__(apply)

        self.width = width
        self.height = height
        self.channels = channels
        self.method = method
        self.mode = mode

    def transform(self, img, rng):
        
        rimg = resize_image(img, self.width, self.height, self.method, self.mode)
        
        if self.channels is None:
            if rimg.format != img.format:
                rimg = rimg.convert(img.format)
            return rimg
        elif self.channels == 1:
            return rimg.convert('gray')
        elif self.channels == 3:
            return rimg.convert('RGB')


from vergeml.utils import VergeMLError, did_you_mean
from vergeml.io import SourcePlugin, source
from vergeml.option import option
from vergeml.sources.image import ImageSource
from vergeml.sources.labeled_image import LabeledImageSource
from vergeml.img import INPUT_PATTERNS, resize_image
import os.path
import numpy as np

# TODO rename image-features

class ImageNetFeatures:

    def __init__(self, args: dict={}):

        self.model = None
        self.preprocess_input = None

        self.variant = args.get('variant')
        self.alpha = args.get('alpha')
        self.size = args.get('size')
        self.output_layer = args.get('output-layer')
        self.architecture = args.get('architecture')
        trainings_dir = args.get('trainings-dir')
        evaluate_args(self.architecture, trainings_dir, self.variant, self.alpha, self.size)
        self.image_size = get_image_size(self.architecture, self.variant, self.size)

@source('image-features', descr='Load Images and convert to feature vectors.', input_patterns=INPUT_PATTERNS)
@option('output-layer', default='last', descr='Index or name of the output layer to use.', type="Union[str,int]")
@option('architecture', default='resnet-50', descr='Name of the CNN to use. Use @name for your own.', type="Union[str,int]")
@option('variant', default='auto', descr='The variant of the CNN.', type=str)
@option('size', default="auto", descr='The input size of the CNN.', type='Union[str, int]')
@option('alpha', default=1.0, descr='MobileNet alpha value.', type=float)
class ImageFeaturesSource(ImageSource, ImageNetFeatures):

    def __init__(self, args: dict={}):
        super().__init__(args)
        ImageNetFeatures.__init__(self, args)

    def transform(self, sample):
        if not self.model:
            if not self.architecture.startswith("@"):
                _, self.preprocess_input, self.model = \
                    get_imagenet_architecture(self.architecture, self.variant, self.size, self.alpha, self.output_layer)
            else:
                self.model = get_custom_architecture(self.architecture, self.trainings_dir, self.output_layer)
                self.preprocess_input = generic_preprocess_input

        x = sample.x
        x = x.convert('RGB')
        x = resize_image(x, self.image_size, self.image_size, 'antialias', 'aspect-fill')
        #x = x.resize((self.image_size, self.image_size))
        x = np.asarray(x)
        x = np.expand_dims(x, axis=0)
        x = self.preprocess_input(x)
        features = self.model.predict(x)
        features = features.flatten()
        sample.x = features
        sample.y = None
        return sample

@source('labeled-image-features', descr='Load labeled Images and convert to feature vectors.', input_patterns=INPUT_PATTERNS)
@option('output-layer', default='last', descr='Index or name of the output layer to use.', type="Union[str,int]")
@option('architecture', default='resnet-50', descr='Name of the CNN to use. Use @name for your own.', type="Union[str,int]")
@option('variant', default='auto', descr='The variant of the CNN.', type=str)
@option('size', default="auto", descr='The size of the CNN.', type='Union[str, int]')
@option('alpha', default=1.0, descr='MobileNet alpha value.', type=float)
class LabeledImageFeaturesSource(LabeledImageSource, ImageNetFeatures):

    def __init__(self, args: dict={}):
        super().__init__(args)
        ImageNetFeatures.__init__(self, args)

    def transform(self, sample):
        if not self.model:
            if not self.architecture.startswith("@"):
                self.preprocess_input = get_preprocess_input(self.architecture)
                self.model = get_imagenet_architecture(self.architecture, self.variant, self.image_size, self.alpha, self.output_layer)
            else:
                # TODO get image size!
                self.model = get_custom_architecture(self.architecture, self.trainings_dir, self.output_layer)
                self.preprocess_input = generic_preprocess_input

        x = sample.x
        # TODO better resize
        x = x.convert('RGB')
        x = resize_image(x, self.image_size, self.image_size, 'antialias', 'aspect-fill')
        # x = x.resize((self.image_size, self.image_size))
        x = np.asarray(x)
        x = np.expand_dims(x, axis=0)
        x = self.preprocess_input(x)
        features = self.model.predict(x)
        features = features.flatten()
        sample.x = features
        sample = super().transform(sample)
        return sample



ARCHITECTURES = (
    'densenet',
    'inception-v3',
    'inception-resnet-v3',
    'mobilenet',
    'mobilenet-v2',
    'nasnet',
    'resnet-50',
    'vgg16',
    'vgg19',
    'xception'
)

DENSENET_VARIANTS = (
    'densenet-121',
    'densenet-169',
    'densenet-201'
)

MOBILENET_SIZES = (
    128,
    160,
    192,
    224
)

MOBILENET_ALPHA_VALUES = (
    0.25,
    0.50,
    0.75,
    1.0
)

MOBILENET_V2_SIZES = (
    96,
    128,
    160,
    192,
    224
)

MOBILENET_V2_ALPHA_VALUES = (
    0.35,
    0.50,
    0.75,
    1.0,
    1.3,
    1.4
)

NASNET_VARIANTS = (
    'large',
    'mobile'
)

def get_custom_architecture(name, trainings_dir, output_layer):
    from keras.models import load_model, Model
    name = name.lstrip("@")
    model = load_model(os.path.join(trainings_dir, name, 'checkpoints', 'model.h5'))
    try:
        if isinstance(output_layer, int):
            layer = model.layers[output_layer]
        else:
            layer = model.get_layer(output_layer)
    except Exception:
        if isinstance(output_layer, int):
            raise VergeMLError(f'output-layer {output_layer} not found - model has only {len(model.layers)} layers.')
        else:
            candidates = list(map(lambda l: l.name, model.layers))
            raise VergeMLError(f'output-layer named {output_layer} not found.',
                               suggestion=did_you_mean(candidates, output_layer))
    model = Model(inputs=model.input, outputs=layer.output)
    return model


def get_imagenet_architecture(architecture, variant, size, alpha, output_layer, include_top=False, weights='imagenet'):
    from keras import applications, Model

    if include_top:
        assert output_layer == 'last'

    if size == 'auto':
        size = get_image_size(architecture, variant, size)

    shape = (size, size, 3)

    if architecture == 'densenet':
        if variant == 'auto':
            variant = 'densenet-121'
        if variant == 'densenet-121':
            model = applications.DenseNet121(weights=weights, include_top=include_top, input_shape=shape)
        elif variant == 'densenet-169':
            model = applications.DenseNet169(weights=weights, include_top=include_top, input_shape=shape)
        elif variant == 'densenet-201':
            model = applications.DenseNet201(weights=weights, include_top=include_top, input_shape=shape)
    elif architecture == 'inception-resnet-v2':
        model = applications.InceptionResNetV2(weights=weights, include_top=include_top, input_shape=shape)
    elif architecture == 'mobilenet':
        model = applications.MobileNet(weights=weights, include_top=include_top, input_shape=shape, alpha=alpha)
    elif architecture == 'mobilenet-v2':
        model = applications.MobileNetV2(weights=weights, include_top=include_top, input_shape=shape, alpha=alpha)
    elif architecture == 'nasnet':
        if variant == 'auto':
            variant = 'large'
        if variant == 'large':
            model = applications.NASNetLarge(weights=weights, include_top=include_top, input_shape=shape)
        else:
            model = applications.NASNetMobile(weights=weights, include_top=include_top, input_shape=shape)
    elif architecture == 'resnet-50':
        model = applications.ResNet50(weights=weights, include_top=include_top, input_shape=shape)
    elif architecture == 'vgg-16':
        model = applications.VGG16(weights=weights, include_top=include_top, input_shape=shape)
    elif architecture == 'vgg-19':
        model = applications.VGG19(weights=weights, include_top=include_top, input_shape=shape)
    elif architecture == 'xception':
        model = applications.Xception(weights=weights, include_top=include_top, input_shape=shape)
    elif architecture == 'inception-v3':
        model = applications.InceptionV3(weights=weights, include_top=include_top, input_shape=shape)

    if output_layer != 'last':
        try:
            if isinstance(output_layer, int):
                layer = model.layers[output_layer]
            else:
                layer = model.get_layer(output_layer)
        except Exception:
            raise VergeMLError('layer not found: {}'.format(output_layer))
        model = Model(inputs=model.input, outputs=layer.output)

    return model

def generic_preprocess_input(input):
    """work around ValueError: output array is read-only when using:
        return mobilenetv2.preprocess_input(input)"""
    input = input / 128
    input = input - 1.
    return input.astype(np.float32)

def get_preprocess_input(architecture):
    from keras import applications
    preprocess_input = {
        'densenet': applications.densenet.preprocess_input,
        'inception-resnet-v2': applications.inception_resnet_v2.preprocess_input,
        'inception-v3': applications.inception_v3.preprocess_input,
        'mobilenet': applications.mobilenet.preprocess_input,
        'mobilenet-v2': generic_preprocess_input,
        'nasnet': applications.nasnet.preprocess_input,
        'resnet-50': applications.resnet50.preprocess_input,
        'vgg-16': applications.vgg16.preprocess_input,
        'vgg-19': applications.vgg19.preprocess_input,
        'xception': applications.xception.preprocess_input
    }
    return preprocess_input[architecture]

def get_decode_predictions(architecture):
    from keras import applications
    decode_predictions = {
        'densenet': applications.densenet.decode_predictions,
        'inception-resnet-v2': applications.inception_resnet_v2.decode_predictions,
        'inception-v3': applications.inception_v3.decode_predictions,
        'mobilenet': applications.mobilenet.decode_predictions,
        'mobilenet-v2': applications.mobilenetv2.decode_predictions,
        'nasnet': applications.nasnet.decode_predictions,
        'resnet-50': applications.resnet50.decode_predictions,
        'vgg-16': applications.vgg16.decode_predictions,
        'vgg-19': applications.vgg19.decode_predictions,
        'xception': applications.xception.decode_predictions
    }
    return decode_predictions[architecture]

def get_image_size(architecture, variant=None, size=None):
    # TODO make this work for @custom-AI
    if architecture == 'densenet':
        image_size = 224
    elif architecture == 'inception-v3':
        image_size = 299
    elif architecture == 'inception-resnet-v2':
        image_size = 299
    elif architecture == 'mobilenet' or architecture == 'mobilenet-v2':
        if size == "auto":
            image_size = 224
        else:
            image_size = size
    elif architecture == 'nasnet':
        if variant == 'large':
            image_size = 331
        else:
            image_size = 224
    elif architecture == 'resnet-50':
        image_size = 224
    elif architecture == 'vgg-16':
        image_size = 224
    elif architecture == 'vgg-19':
        image_size = 224
    elif architecture == 'xception':
        image_size = 299
    return image_size

def evaluate_args(architecture, trainings_dir, variant, alpha, size):

    if not architecture.startswith('@') and not architecture in ARCHITECTURES:
        raise VergeMLError("Unknown CNN: {}".format(architecture))
    elif architecture.startswith('@'):
        name = architecture.lstrip('@')
        path = os.path.join(trainings_dir, name, 'checkpoints', 'model.h5')
        if not os.path.isfile(path):
            raise VergeMLError("Unknown CNN: {}".format(architecture))

    if architecture == 'densenet':
        if variant == 'auto':
            variant = DENSENET_VARIANTS[0]
        if not variant in DENSENET_VARIANTS:
            raise VergeMLError("Invalid densenet variant: {}".format(variant))

    elif architecture == 'mobilenet':
        if size not in MOBILENET_SIZES and size != "auto":
            raise VergeMLError("Invalid mobilenet size: {}".format(size))
        if alpha not in MOBILENET_ALPHA_VALUES:
            raise VergeMLError("Invalid alpha value: {}".format(alpha))

    elif architecture == 'mobilenet-v2':
        if size not in MOBILENET_V2_SIZES and size != "auto":
            raise VergeMLError("Invalid mobilenet size: {}".format(size))

        if alpha not in MOBILENET_V2_ALPHA_VALUES:
            raise VergeMLError("Invalid alpha value: {}".format(alpha))

    elif architecture == 'nasnet':
        if not variant in NASNET_VARIANTS and variant != 'auto':
            raise VergeMLError("Invalid nasnet variant: {}".format(variant))


"""VergeML.
"""
__version__ = "0.1.4"


from vergeml.utils import VergeMLError, SPLITS
from vergeml.command import command, train, predict, CommandPlugin
from vergeml.option import option
from vergeml.model import ModelPlugin, model
from vergeml.data import Data



from vergeml.command import command, CommandPlugin
from vergeml.option import option
from vergeml.utils import VergeMLError
import sys
import os.path

@command('new', descr="Create a new VergeML project.")
@option('<project-name>', type=str)
class NewCommand(CommandPlugin):

    def __call__(self, args, env):
        dest = args['<project-name>']
        if not env.model_plugin:
            template = "# model:\n#   name: <name of your model>\n"
        else:
            template = env.model_plugin.project_file_template()

        if not template.endswith("\n"):
            template = template + "\n"

        if os.path.exists(dest):
            raise VergeMLError("Directory already exists: {}".format(dest))

        os.makedirs(dest)
        os.makedirs(os.path.join(dest, "samples"))
        with open(os.path.join(dest, "vergeml.yaml"), "w") as f:
            f.write(template)
        print("Created new project: {}".format(dest))



# pylint: disable=C0111
from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    LONG_DESCRIPTION = fh.read()


setup(
    name="vergeml",
    version="0.1.5",
    author="Markus Ecker",
    author_email="markus.ecker@gmail.com",
    description="Machine Learning Environment",
    license="MIT",
    keywords="ai deep learning",
    url="http://github.com/vergeml/vergeml",
    packages=find_packages(),
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Topic :: Utilities",
        "License :: OSI Approved :: MIT License",
    ],
    include_package_data=True,
    install_requires=[
        'numpy',
        'pyyaml',
        'lz4',
        'scikit-learn',
        # install this for the first release only - in a subsequent release dependencies should be
        # installed on demand
        'Pillow',
        'keras',
        'waitress',
        'matplotlib'
    ],
    # preprocess=vergeml.commands.preprocess:preprocess
    # mnist=vergeml.datasets.mnist:download
    # fashion-mnist=vergeml.datasets.fashion_mnist:download
    # cifar-10=vergeml.datasets.cifar_10:download
    # cifar-100=vergeml.datasets.cifar_100:download
    # celeba=vergeml.datasets.celeba:download
    # svhn=vergeml.datasets.svhn:download

    # dogs=vergeml.datasets.dogs:download
    # unique-objects=vergeml.datasets.unique_objects:download
    entry_points="""
    [vergeml.cmd]
    help=vergeml.commands.help:HelpCommand
    new=vergeml.commands.new:NewCommand
    list=vergeml.commands.ls:ListCommand
    download=vergeml.commands.download:DownloadCommand
    run=vergeml.commands.run:RunCommand
    plot=vergeml.commands.plot:PlotCommand
    preprocess=vergeml.commands.preprocess:PreprocessCommand

    [vergeml.download]
    cats-and-dogs=vergeml.datasets.cats_and_dogs:CatsAndDogsDataset
    ham10000=vergeml.datasets.ham10000:Ham10KDataset
    unique-objects=vergeml.datasets.unique_objects:UniqueObjectsDataset
    dogs=vergeml.datasets.dogs:DogsDataset

    [vergeml.run]
    tensorboard=vergeml.services.tensorboard:TensorboardService
    rest=vergeml.services.rest:RestService

    [vergeml.plot]
    roc=vergeml.plots.roc:ROCPlot
    confusion-matrix=vergeml.plots.confusion_matrix:ConfusionMatrixPlot
    pr=vergeml.plots.pr:PRPlot

    [vergeml.operation]
    augment=vergeml.operations.augment:AugmentOperation
    resize=vergeml.operations.resize:ResizeOperation[Pillow]
    grayscale=vergeml.operations.grayscale:GrayscaleOperation[Pillow]
    rgb=vergeml.operations.rgb:RGBOperation[Pillow]
    crop=vergeml.operations.crop:CropOperation[Pillow]
    random-crop=vergeml.operations.random_crop:RandomCropOperation[Pillow]
    flip-horizontal=vergeml.operations.flip_horizontal:FlipHorizontalOperation[Pillow]
    flip-vertical=vergeml.operations.flip_vertical:FlipVerticalOperation[Pillow]

    [vergeml.io]
    image=vergeml.sources.image:ImageSource[Pillow]
    labeled-image=vergeml.sources.labeled_image:LabeledImageSource[Pillow]
    labeled-image-features=vergeml.sources.features:LabeledImageFeaturesSource[Pillow]
    image-features=vergeml.sources.features:ImageFeaturesSource[Pillow]
    mnist=vergeml.sources.mnist:plugin[Pillow]

    [vergeml.model]
    imagenet=vergeml.models.imagenet:ImageNetModelPlugin

    [console_scripts]
    ml = vergeml.__main__:main
    """
)



"""Utility functions
"""
import inspect
import textwrap
from collections import namedtuple
import re
import os

SPLITS = ('train', 'val', 'test')


class VergeMLError(Exception):
    """System error.
    """

    def __init__(self, # pylint: disable=R0913
                 message, suggestion=None, help_topic=None, hint_type=None, hint_key=None):
        super().__init__(message)
        self.suggestion = suggestion
        self.message = message
        self.hint_type = hint_type
        self.hint_key = hint_key
        self.help_topic = help_topic

    def __str__(self):
        if self.suggestion:
            if len(self.message + self.suggestion) < 80:
                return self.message + " " + self.suggestion
            return self.message + "\n" + self.suggestion
        return self.message


def wrap_text(text):
    """Wrap text to be readable in the terminal.
    """
    # TODO check terminal width
    res = []
    for para in text.split("\n\n"):
        if para.splitlines()[0].strip().endswith(":"):
            res.append(para)
        else:
            res.append(textwrap.fill(para, drop_whitespace=True, fix_sentence_endings=True))
    return "\n\n".join(res)


_Intro = namedtuple('_Intro', ['args', 'defaults', 'types'])


def introspect(call):
    """Introspect a function call.
    """
    spec = inspect.getfullargspec(call)
    args = spec.args
    defaults = dict(zip(reversed(spec.args), reversed(spec.defaults or [])))
    types = spec.annotations
    return _Intro(args, defaults, types)


# taken from here: https://www.python-course.eu/levenshtein_distance.php
def _iterative_levenshtein(source, targ):
    """
        iterative_levenshtein(source, targ) -> ldist
        ldist is the Levenshtein distance between the strings
        source and targ.
        For all i and j, dist[i,j] will contain the Levenshtein
        distance between the first i characters of source and the
        first j characters of targ
    """
    rows = len(source)+1
    cols = len(targ)+1
    dist = [[0 for x in range(cols)] for x in range(rows)]
    # source prefixes can be transformed into empty strings
    # by deletions:
    for i in range(1, rows):
        dist[i][0] = i
    # target prefixes can be created from an empty source string
    # by inserting the characters
    for i in range(1, cols):
        dist[0][i] = i

    row, col = None, None

    for col in range(1, cols):
        for row in range(1, rows):
            if source[row-1] == targ[col-1]:
                cost = 0
            else:
                cost = 1
            dist[row][col] = min(dist[row-1][col] + 1,      # deletion
                                 dist[row][col-1] + 1,      # insertion
                                 dist[row-1][col-1] + cost) # substitution

    assert row and col
    return dist[row][col]


def did_you_mean(candidates, value, fmt="'{}'"):
    """In case of a misspelling, return possible candidates.
    """
    candidates = list(candidates)
    names = list(sorted(map(lambda n: (_iterative_levenshtein(value, n), n), candidates)))
    names = list(filter(lambda dn: dn[0] <= 2, names))
    return 'Did you mean ' + fmt.format(names[0][1]) + '?' if names else None


def dict_set_path(dic, path, value):
    """Set the value of a dict using path syntax.
    """
    cur = dic
    path = path.split(".")
    for key in path[:-1]:
        cur = cur.setdefault(key, {})
    cur[path[-1]] = value

def dict_del_path(dic, path):
    """Delete a value from a dict using path syntax.
    """
    if isinstance(path, str):
        path = path.split(".")
    if len(path) == 1:
        del[dic[path[0]]]
    else:
        pat, *rest = path
        dict_del_path(dic[pat], rest)
        if not dic[pat]:
            del dic[pat]

def dict_has_path(dic, path):
    """Check if a dict contains a value using path syntax.
    """
    cur = dic
    for pat in path.split("."):
        if isinstance(cur, dict) and pat in cur:
            cur = cur[pat]
        else:
            return False
    return True

_DEFAULT = object()
def dict_get_path(dic, path, default=_DEFAULT):
    """Get the value of a dict using path syntax.
    """
    cur = dic
    for pat in path.split("."):
        if isinstance(cur, dict) and pat in cur:
            cur = cur[pat]
        elif default != _DEFAULT:
            return default
        else:
            raise KeyError(path)
    return cur

def dict_merge(dict1, dict2):
    """Merge two dicts.
    """
    if not isinstance(dict1, dict) or not isinstance(dict2, dict):
        return dict2
    for k in dict2:
        if k in dict1:
            dict1[k] = dict_merge(dict1[k], dict2[k])
        else:
            dict1[k] = dict2[k]
    return dict1

def dict_paths(dic, path=None):
    """Get paths in a dict.
    """
    res = []
    if path:
        if not dict_has_path(dic, path):
            return res
        value = dict_get_path(dic, path)
    else:
        value = dic
    if not isinstance(dic, dict):
        return res
    def _collect_path(dic, path):
        for k, val in dic.items():
            npath = f"{path}.{k}" if path is not None else k
            if isinstance(val, dict):
                _collect_path(val, npath)
            else:
                res.append(npath)
    _collect_path(value, path)
    return res


def parse_trained_models(argv):
    """Parse @syntax for specifying trained models on the command line.
    """
    names = []
    for part in argv:
        if re.match("^@[a-zA-Z0-9_-]+$", part):
            names.append(part[1:])
        else:
            break
    rest = argv[len(names):]
    return names, rest

def parse_split(value):
    """Decodes the split value.

    Returns a tuple (type, value) where type is either perc, num or dir set.
    """
    assert isinstance(value, (int, str))

    if isinstance(value, int):
        return ('num', value)
    if value.endswith("%"):
        return ('perc', float(value.rstrip("%").strip()))
    if value.isdigit():
        return ('num', int(value))
    return ('dir', value)

def format_info_text(text, indent=0, width=70):
    """Return text formatted for readability.
    """
    text = text.strip("\n")
    res = []
    for line in text.splitlines():
        if line.startswith("  "):
            res.append(line)
        elif line.strip() == "":
            res.append(line)
        else:
            res.extend(textwrap.wrap(line, width=width-indent))
    if indent:
        indstr = str(' ' * indent)
        res = list(map(lambda l: indstr + l, res))
    return "\n".join(res)

if os.name == 'nt':
    def xlink(src, dst):
        """Cross platform file links.
        """
        os.link(src, dst)
else:
    def xlink(src, dst):
        """Cross platform file links.
        """
        # use symlink on Unix
        os.symlink(src, dst)



from vergeml.utils import download_files
from vergeml.utils import Error
import shutil
import os.path
import zipfile
import tarfile


def download(env):
    """The MNIST database of handwritten digits.

The MNIST database of handwritten digits has a training set of
60,000 examples, and a test set of 10,000 examples. It is a
subset of a larger set available from NIST. The digits have been
size-normalized and centered in a fixed-size image.

It is a good database for people who want to try learning
techniques and pattern recognition methods on real-world data
while spending minimal efforts on preprocessing and formatting.

Authors:
  Yann LeCun, Courant Institute, NYU
  Corinna Cortes, Google Labs, New York
  Christopher J.C. Burges, Microsoft Research, Redmond

For more information visit:http://yann.lecun.com/exdb/mnist/"""

    urls = ["http://yann.lecun.com/exdb/mnist/train-images-idx3-ubyte.gz",
            "http://yann.lecun.com/exdb/mnist/train-labels-idx1-ubyte.gz",
            "http://yann.lecun.com/exdb/mnist/t10k-images-idx3-ubyte.gz",
            "http://yann.lecun.com/exdb/mnist/t10k-labels-idx1-ubyte.gz"]

    samples_dir = env.get('base.samples_dir')
    print("Downloading mnist to {}.".format(samples_dir))

    src_dir = download_files(urls, dir=env.get('base.cache_dir'))

    for file in ("train-images-idx3-ubyte.gz", "train-labels-idx1-ubyte.gz",
                 "t10k-images-idx3-ubyte.gz", "t10k-labels-idx1-ubyte.gz"):
        shutil.copy(os.path.join(src_dir, file), samples_dir)

    shutil.rmtree(src_dir)

    print("Finished downloading mnist.")


download.__info__ = [
    ('Samples', '60K'),
    ('Test Samples', '10K'),
    ('Type', 'Labeled Images'),
    ('Resolution', '28x28 grayscale'),
    ('Size', '12 MB')
]



import os.path
import math
from PIL import Image
from PIL.Image import Image as ImageType


INPUT_PATTERNS = ["**/*.jpg", "**/*.jpeg", "**/*.png", "**/*.bmp"]
RESIZE_METHODS = ('nearest', 'box', 'bilinear', 'hamming', 'bicubic', 'lanczos', 'antialias')
# Apple:
# https://developer.apple.com/documentation/uikit/uiview/contentmode
# Keras:
# 'constant': kkkkkkkk|abcd|kkkkkkkk (cval=k)
# 'nearest': aaaaaaaa|abcd|dddddddd
# 'reflect': abcddcba|abcd|dcbaabcd
# 'wrap': abcdabcd|abcd|abcdabcd

RESIZE_MODES = ('fill', 'aspect-fill', 'aspect-fit')

def fixext(path, img):
    """Change the format of files with the wrong extension."""
    path, ext = os.path.splitext(path)

    if img.format:
        return path + "." + img.format.lower()
    elif img.mode == 'RGBA':
        return path + ".png"
    elif ext.lower() not in [".jpg", ".jpeg", ".png", ".bmp"]:
        return path + ".png"
    else:
        return path + ext


def open_image(path):
    """Open image at path.

    PIL lazily opens the image, which can lead to a 'too many open files' error.
    This workaround reads the file into memory immediately."""
    img1 = Image.open(path)
    img2 = img1.copy()
    img1.close()
    return img2

def resize_image(img, width, height, method, mode, bg_color=(0, 0, 0, 0)):
    # Some code from:
    # https://github.com/charlesthk/python-resize-image/blob/master/resizeimage/resizeimage.py
    # Thank you!
    img = img.copy()
    pil_method = getattr(Image, method.upper())

    if mode == 'fill':
        img = img.resize((width, height), pil_method)
    elif mode == 'aspect-fill':
        w,h = img.size
        ratio = max(width / w, height / h)
        nsize = (int(math.ceil(w * ratio)), int(math.ceil(h * ratio)))
        img = img.resize(nsize, pil_method)
        w,h = img.size
        left = (w - width) / 2
        top = (h - height) / 2
        right = w - left
        bottom = h - top
        rect = (int(math.ceil(x)) for x in (left, top, right, bottom))
        img = img.crop(rect)
    elif mode == 'aspect-fit':
        img.thumbnail((width, height), pil_method)
        background = Image.new('RGBA', (width, height), bg_color)
        img_position = (
            int(math.ceil((width - img.width) / 2)),
            int(math.ceil((height - img.height) / 2))
        )
        background.paste(img, img_position)
        img = background.convert('RGB')
    
    return img


from vergeml.utils import download_files
from vergeml.utils import Error
import shutil
import os.path
import zipfile
import tarfile


def download(env):
    """A MNIST-like fashion product database.

Fashion-MNIST is a dataset of Zalando's article images—
consisting of a training set of 60,000 examples and a test set
of 10,000 examples. Each example is a 28x28 grayscale image,
associated with a label from 10 classes. We intend Fashion-MNIST
to serve as a direct drop-in replacement for the original MNIST
dataset for benchmarking machine learning algorithms. It shares
the same image size and structure of training and testing splits.

Authors:
  Han Xiao
  Kashif Rasul
  Roland Vollgraf

See also:
  Fashion-MNIST: a Novel Image Dataset for Benchmarking Machine
  Learning Algorithms. arXiv:1708.07747
  https://arxiv.org/abs/1708.07747

For more information visit:
https://github.com/zalandoresearch/fashion-mnist"""

    urls = ["http://fashion-mnist.s3-website.eu-central-1.amazonaws.com/train-images-idx3-ubyte.gz",
            "http://fashion-mnist.s3-website.eu-central-1.amazonaws.com/train-labels-idx1-ubyte.gz",
            "http://fashion-mnist.s3-website.eu-central-1.amazonaws.com/t10k-images-idx3-ubyte.gz",
            "http://fashion-mnist.s3-website.eu-central-1.amazonaws.com/t10k-labels-idx1-ubyte.gz"]

    samples_dir = env.get('base.samples_dir')
    print("Downloading fashion mnist to {}.".format(samples_dir))

    src_dir = download_files(urls, dir=env.get('base.cache_dir'))

    for file in ("train-images-idx3-ubyte.gz", "train-labels-idx1-ubyte.gz",
                 "t10k-images-idx3-ubyte.gz", "t10k-labels-idx1-ubyte.gz"):
        shutil.copy(os.path.join(src_dir, file), samples_dir)

    shutil.rmtree(src_dir)

    print("Finished downloading fashion mnist.")


download.__info__ = [
    ('Samples', '60K'),
    ('Test Samples', '10K'),
    ('Type', 'Labeled Images'),
    ('Resolution', '28x28 grayscale'),
    ('Size', '30 MB')
]



"""Run command.
"""
from vergeml.command import command, CommandPlugin
from vergeml.option import option

# pylint: disable=R0903

@command('run', descr="Run a service or external program.")
@option('service', type=str, subcommand='vergeml.run')
class RunCommand(CommandPlugin):
    """Run a service or external program.
    """

    def __call__(self, args, env):
        plugin = self.plugins.get('vergeml.run', args['service'])(args['service'])
        plugin(args, env)



from vergeml.utils import VergeMLError
from vergeml.dataset import DatasetPlugin, dataset
from vergeml.display import DISPLAY
import shutil
import os.path
import zipfile
import tarfile


_LONG_DESCR = """2400 objects of distinct categories.

Credits:
  Brady, T. F., Konkle, T., Alvarez, G. A. and Oliva, A. (2008). 
  Visual long-term memory has a massive storage capacity for 
  object details. Proceedings of the National Academy of 
  Sciences, USA, 105 (38), 14325-14329.
  http://cvcl.mit.edu/MM/pdfs/BradyKonkleAlvarezOliva2008.pdf

For more information visit: 
http://cvcl.mit.edu/MM/uniqueObjects.html""".strip()

_INFO = [
    ['Samples', '2400'],
    ['Type', 'Images'],
    ['Resolution', '256x256 RGB'],
    ['Size', '67.8 MB']
]


_LONG_DESCR += "\n\n" + DISPLAY.table(_INFO, separate='none').getvalue()

_URL = "http://cvcl.mit.edu/MM/downloads/ObjectsAll.zip"

@dataset('unique-objects', descr="2400 objects of distinct categories.", long_descr=_LONG_DESCR)
class UniqueObjectsDataset(DatasetPlugin):

    def __call__(self, args, env):
        samples_dir = env.get('samples-dir')

        print("Downloading unique objects to {}.".format(samples_dir))

        src_dir = self.download_files([_URL], env=env, dir=env.get('cache-dir'))
        path = os.path.join(src_dir, "ObjectsAll.zip")

        zipf = zipfile.ZipFile(path, 'r')
        zipf.extractall(src_dir)
        zipf.close()

        for file in os.listdir(os.path.join(src_dir, "OBJECTSALL")):
            shutil.copy(os.path.join(src_dir, "OBJECTSALL", file), samples_dir)

        shutil.rmtree(src_dir)

        print("Finished downloading unique objects.")



"""Code for running from the command line.
"""

import sys
import os.path
import logging
import getopt
import re
from copy import deepcopy, copy

from vergeml import __version__
from vergeml.utils import VergeMLError, parse_trained_models
from vergeml.env import Environment
from vergeml.plugins import PLUGINS
from vergeml.command import Command
from vergeml.commands.help import HelpCommand
from vergeml.utils import did_you_mean
from vergeml.libraries import KerasLibrary, TensorFlowLibrary, TorchLibrary
from vergeml.libraries import NumPyLibrary, PythonInterpreter, CudaLibrary


def _parsebase(argv):
    """Parse until the second part of the command.
    """
    shortopts = 'vf:m:' # version, file, model
    longopts = ['version', 'file=', 'model=', 'samples-dir=', 'test-split=', 'val-split=',
                'cache-dir=', 'random-seed=', 'trainings-dir=', 'project-dir=',
                'cache=', 'device=', 'device-memory=']

    args, rest = getopt.getopt(argv, shortopts, longopts)

    args = dict(args)
    # don't match prefix
    for opt in map(lambda s: s.rstrip("="), longopts):
        # pylint: disable=W0640
        if ''f'--{opt}' in args and not any(map(lambda a: a.startswith('--' + opt), argv)):
            # find the key that does not match
            keys = map(lambda a: a.split("=")[0].lstrip("-"), argv)
            keys = list(filter(lambda k: k in opt, keys))
            if keys:
                raise getopt.GetoptError('Invalid key', opt='--' + keys[0])
            else:
                raise getopt.GetoptError('Invalid key')

    # convert from short to long names
    for sht, lng in (('-v', '--version'), ('-m', '--model'), ('-f', '--file')):
        if sht in args:
            args[lng] = args[sht]
            del args[sht]

    args = {k.strip('-'):v for k, v in args.items()}

    return args, rest


def _env_from_args(args, trained_model, plugins=PLUGINS):
    args = deepcopy(args)
    # replace hyphen with underscore for python
    args = {k.replace('-', '_'):v for k, v in args.items()}

    if trained_model:
        args['trained_model'] = trained_model

    args['is_global_instance'] = True
    args['plugins'] = plugins

    env = Environment(**args)

    return env

def _prepare_args(args):
    """Prepare args by appending the project dir and setting defaults.
    """
    args = deepcopy(args)
    project_dir = args.get('project-dir', '')

    if not 'file' in args:
        default_file = os.path.join(project_dir, "vergeml.yaml")
        if os.path.exists(default_file):
            args['file'] = default_file

    if 'file' in args:
        args['project-file'] = args['file']
        del args['file']

    if 'random-seed' in args:
        try:
            args['random-seed'] = int(args['random-seed'])
        except ValueError:
            raise VergeMLError("Invalid value for --random-seed.",
                               "--random-seed must be an integer value.",
                               ('value', 'random-seed'))

    cache_opts = ('none', 'disk', 'mem', 'disk-in', 'mem-in')

    if 'cache' in args:
        if args['cache'] not in cache_opts:
            raise VergeMLError("Invalid value for --cache.",
                               "Must be one of: " + ", ".join(cache_opts),
                               help_topic='cache')

    if 'device' in args:
        if not re.match(r"^(gpu:[0-9]+|gpu|cpu|auto)", args['device']):
            raise VergeMLError("Invalid value for --device.",
                               "Please specify a valid device, e.g gpu:0 or cpu.",
                               help_topic='device')

    if 'device-memory' in args:
        if not re.match(r"(([1-9]?[0-9]|100)%|(0\.[0-9]+)|1\.0)|auto", args['device-memory']):
            raise VergeMLError("Invalid value for --device-memory.",
                               "Please specify device memory as a percentage, e.g. 100%.",
                               help_topic='device')

    return args

_VERGEML_OPTION_NAMES = {
    'version', 'file', 'model', 'samples-dir', 'val-split', 'test-split', 'cache-dir',
    'random-seed', 'trainings-dir', 'project-dir', 'cache', 'device', 'device-memory'
}

def _forgive_wrong_option_order(argv):
    first_part = []
    second_part = []
    rest = copy(argv)

    while rest:
        arg = rest.pop(0)

        if arg.startswith("--"):
            argname = arg.lstrip("--")
            if "=" in argname:
                argname = argname.split("=")[0]
            is_vergeml_opt = bool(argname in _VERGEML_OPTION_NAMES)
            lst = (first_part if is_vergeml_opt else second_part)

            if arg.endswith("=") or not "=" in arg:
                if not rest:
                    # give up
                    second_part.append(arg)
                else:
                    lst.append(arg)
                    lst.append(rest.pop(0))
            else:
                lst.append(arg)

        else:
            second_part.append(arg)

    return first_part + second_part

def run(argv, plugins=PLUGINS):
    """Run from command line.
    """
    try:
        argv = _forgive_wrong_option_order(argv)
        args, rest = _parsebase(argv)
    except getopt.GetoptError as err:
        if err.opt:
            opt = err.opt.lstrip("-")
            dashes = '-' if len(opt) == 1 else '--'
            raise VergeMLError(f"Invalid option {dashes}{opt}.", help_topic='options')
        else:
            raise VergeMLError(f"Invalid option.", help_topic='options')

    if 'version' in args:
        print_version()
        exit()

    args = _prepare_args(args)
    ai_names, after_names = parse_trained_models(rest)

    AI = next(iter(ai_names), None)

    env = _env_from_args(args, AI, plugins=plugins)

    if after_names:
        cmdname = after_names.pop(0)
    else:
        cmdname = 'help'
        rest = ['help']

    if ":" in cmdname:
        cmdname = cmdname.split(":")[0]

    plugin = None
    cmd_plugin = plugins.get('vergeml.cmd', cmdname)
    if cmd_plugin:
        plugin = cmd_plugin(cmdname, plugins=plugins)
    elif env.model_plugin:
        for model_fn in Command.find_functions(env.model_plugin):
            if cmdname == Command.discover(model_fn).name:
                plugin = model_fn
                break

    if not plugin:
        # collect all possible command names
        command_names = set(plugins.keys('vergeml.cmd'))
        if env.model_plugin:
            model_commands = set(map(lambda f:Command.discover(f).name, Command.find_functions(env.model_plugin)))
            command_names.update(model_commands)

        raise VergeMLError(f"Invalid command '{cmdname}'.",
                           suggestion=did_you_mean(command_names, cmdname),
                           help_topic='*help*')
    try:
        cmd = Command.discover(plugin)
        assert cmd
        args = cmd.parse(rest)

        try:
            # return the result for unit testing
            return plugin(args, env)
        finally:
            if env.training is not None:
                env.cancel_training()

    except KeyboardInterrupt:
        # silence the stacktrace
        pass


def print_version():
    """Print VergeML version and versions of various libraries used.
    """
    print("----------------")
    print(f"VergeML {__version__}")
    print("----------------")
    print("")
    print("Installed Libraries:")
    print("")
    sys.stdout.flush()
    for lib, label in [(PythonInterpreter, 'Python'),
                       (NumPyLibrary, 'Numpy'),
                       (TensorFlowLibrary, 'TensorFlow'),
                       (KerasLibrary, 'Keras'),
                       (TorchLibrary, 'PyTorch'),
                       (CudaLibrary, 'CUDA')]:
        if lib.is_installed():
            version = lib.version()
            print(f"  {label} {version}")
            sys.stdout.flush()
    if CudaLibrary.is_installed():
        print("")
        devices = CudaLibrary.devices_info()
        if not devices:
            print("No CUDA devices found.")
        else:
            print("CUDA devices:")
            for device in devices:
                print("")
                print("\n".join(map(lambda l: "  " + l, device.splitlines())))


def _configure_logging(level=logging.INFO):
    logging.addLevelName(logging.DEBUG, 'Debug:')
    logging.addLevelName(logging.INFO, 'Info:')
    logging.addLevelName(logging.WARNING, 'Warning!')
    logging.addLevelName(logging.CRITICAL, 'Critical!')
    logging.addLevelName(logging.ERROR, 'Error!')

    logging.basicConfig(format='%(levelname)s %(message)s', level=logging.INFO)

    if not sys.warnoptions:
        import warnings
        warnings.simplefilter("ignore")
        # TODO hack to get rid of deprecation warning that appeared allthough filters
        # are set to ignore. Is there a more sane way?
        warnings.warn = lambda *args, **kwargs: None

def main(argv=None, plugins=PLUGINS):
    if argv is None:
        argv = sys.argv[1:]
    _configure_logging()
    try:
        run(argv, plugins=plugins)

    except VergeMLError as e:
        # NOTE- when the error is encountered before the environment is created, it will be empty.
        from vergeml.env import ENV
        # in case there is an error with the config file, but the user just says 'ml help <topic>'
        # where topic is the topic suggested by VergeML, try to display the help message
        # instead of the error the user is experiencing
        if ["help", e.help_topic] == argv:
            help = HelpCommand('help')
            from vergeml.env import ENV
            print(help.get_help(ENV, e.help_topic))
        else:
            # display the error. Can't use logging.error because for an unknown reason pytest does not
            # capture stderr when using logging, so fall back to print
            err_string = str(e).strip()
            print("Error! " + err_string, file=sys.stderr)
            # find all command topics
            commands = list(plugins.keys('vergeml.cmd'))
            if ENV and ENV.model_plugin:
                fns = Command.find_functions(ENV.model_plugin)
                mcommands = list(map(lambda f: Command.discover(f).name, fns))
                commands.extend(mcommands)
            # if the error is just one line and there is command help available, display the help message too.
            if e.help_topic and len(err_string.splitlines()) == 1 and e.help_topic in plugins.keys('vergeml.cmd'):
                print("")
                help = HelpCommand('help')
                help_topic = "" if e.help_topic == "*help*" else e.help_topic
                print(help.get_help(ENV, help_topic, short=True))
            # else just hint at the help topic
            elif e.help_topic:
                print("",  file=sys.stderr)
                help_topic = "" if e.help_topic == "*help*" else " " + e.help_topic
                print(f"See 'ml help" + help_topic + "'.", file=sys.stderr)

    except Exception as err: # pylint: disable=W0703
        if err.__class__.__name__ == 'ResourceExhaustedError':
            print("Error! Your GPU ran out of memory.")
            print("Try lowering resource usage by decreasing model parameters such as batch size.")
        else:
            raise err

if __name__ == "__main__":
    main()